# package
from financebuddy.parserconfig.integrations.dataframe import DataframeParserConfig, DataframeParserConfigSettings
from financebuddy.parserconfig.models import ParserExtension


class CSVParserConfigSettings(DataframeParserConfigSettings):
    delimiter: str | None = ","


class CSVParserConfig(DataframeParserConfig):
    type: ParserExtension = ParserExtension.CSV
    settings: CSVParserConfigSettings



# stdlib
from enum import StrEnum

# third party
from pydantic import BaseModel

# package
from financebuddy.report.models import ReportTransaction


class ExporterFormat(StrEnum):
    UNKNOWN = "unknown"
    FINANCEBUDDY = "financebuddy"


class ExporterExtension(StrEnum):
    UNKNOWN = "unknown"
    JSON = "json"
    CSV = "csv"


class ExportTransaction(ReportTransaction):
    pass


class ExportReport(BaseModel):
    transactions: list[ExportTransaction]



# stdlib
import datetime
import json
import os
import tempfile

# package
from financebuddy.report.models import Report


def load_report(path: str) -> Report:
    with open(path, "r") as fp:
        content = json.load(fp)
        report = Report.model_validate(content)
        return report


def dump_report(report: Report) -> str:
    directory = tempfile.gettempdir()
    timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
    filename = f"financebuddy_report_{timestamp}.json"
    path = os.path.join(directory, filename)
    with open(path, "w") as fp:
        data = json.loads(report.model_dump_json())
        json.dump(data, fp, indent=4)
    return path



# stdlib
import datetime
import os
import tempfile

# third party
import pytest

# package
from financebuddy.export.integrations.base import Exporter
from financebuddy.report.models import Report


def test_init():
    assert Exporter.format == "unknown"
    assert Exporter.extension == "unknown"
    Exporter()


def test_report_to_export_report(fix_report: Report):
    export_report = Exporter.report_to_export_report(fix_report)
    assert len(export_report.transactions) == 1


def test_build_export_report_path(freezer, tmpdir, monkeypatch):
    monkeypatch.setattr(tempfile, "gettempdir", lambda: str(tmpdir))
    datetime_str = "1970-01-01T00:00:00.000Z"
    freezer.move_to(datetime_str)
    timestamp = datetime.datetime.now(tz=datetime.timezone.utc).timestamp()

    exported = Exporter()
    path = exported.build_export_report_path()
    dirname, basename = os.path.split(path)
    assert dirname == str(tmpdir)
    assert basename == f"financebuddy_export_unknown_{timestamp}.unknown"


def test_dump_report():
    exporter = Exporter()
    with pytest.raises(NotImplementedError):
        exporter.dump_report(object())



# stdlib
import json

# package
from financebuddy.export.integrations.base import Exporter
from financebuddy.export.models import ExporterExtension, ExporterFormat, ExportReport


class FinanceBuddyJSONExporter(Exporter):
    format: ExporterFormat = ExporterFormat.FINANCEBUDDY
    extension: ExporterExtension = ExporterExtension.JSON

    def dump_report(self, export_report: ExportReport) -> str:
        path = self.build_export_report_path()
        with open(path, "w") as fp:
            data = json.loads(export_report.model_dump_json())
            json.dump(data, fp, indent=4)
        return path



# package
from financebuddy.parserconfig.integrations.csv import CSVParserConfig, CSVParserConfigSettings
from financebuddy.parserconfig.models import ParserExtension


def test_init_config_settings():
    settings = CSVParserConfigSettings(
        field_map={},
    )
    assert settings.field_map == {}
    assert settings.date_format is None
    assert settings.currency is None
    assert settings.delimiter == ","


def test_init_config():
    config = CSVParserConfig(
        path="",
        format="unknown",
        extension=ParserExtension.UNKNOWN,
        settings=CSVParserConfigSettings(field_map={}),
    )
    assert config.type == "csv"



# third party
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parser import factory
from financebuddy.parser.integrations.base import Parser
from financebuddy.parser.integrations.csv import CSVParser
from financebuddy.parserconfig.models import ParserConfig, ParserExtension


@pytest.mark.parametrize(
    ["extension", "ParserType"],
    [
        (ParserExtension.CSV, CSVParser),
    ],
)
def test_get_parser_type(
    extension: ParserExtension,
    ParserType: type[Parser],
):
    assert factory.get_parser_type(extension) == ParserType


def test_get_parser_type_raise_FinanceBuddyException():
    with pytest.raises(FinanceBuddyException):
        factory.get_parser_type("?")


def test_get_parser(fix_parser_config: ParserConfig):
    fix_parser_config.extension = ParserExtension.CSV
    parser = factory.get_parser(fix_parser_config)
    assert isinstance(parser, CSVParser)



# stdlib
import datetime

# third party
import pandas as pd
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parser.integrations.dataframe import DataframeParser, pandas_serialize_json
from financebuddy.report.models import ReportTransactionField


@pytest.mark.parametrize(
    ["value", "expected"],
    [
        ("foo", "foo"),
        (pd.Timestamp(datetime.datetime(1970, 1, 1)), datetime.datetime(1970, 1, 1).isoformat()),
        (1, "1"),
        (1.0, "1.0"),
    ],
)
def test_pandas_serialize_json(value, expected):
    assert pandas_serialize_json(value) == expected


def test_read_file(fix_parser_dataframe: DataframeParser):
    with pytest.raises(NotImplementedError):
        fix_parser_dataframe.read_file("")


def test_row_to_transaction(fix_parser_dataframe: DataframeParser):
    date = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    row = [
        " foo ",
        " bar ",
        date.isoformat(),
        "1.00",
        "EUR",
    ]
    fix_parser_dataframe.config.settings.field_map = {
        ReportTransactionField.SRC_ACCOUNT_NUMBER: 0,
        ReportTransactionField.DST_ACCOUNT_NUMBER: 1,
        ReportTransactionField.DATE: 2,
        ReportTransactionField.VALUE: 3,
        ReportTransactionField.CURRENCY: 4,
    }
    transaction = fix_parser_dataframe.row_to_transaction(row)
    assert transaction.src_account_number == "foo"
    assert transaction.dst_account_number == "bar"
    assert transaction.date == date
    assert transaction.value == 100
    assert transaction.display == "1.00"
    assert transaction.currency == "EUR"


def test_test_row_to_transaction_fail_no_currency(fix_parser_dataframe: DataframeParser):
    row: list[str] = ["1.00"]
    fix_parser_dataframe.config.settings.field_map = {ReportTransactionField.VALUE: 0}
    with pytest.raises(FinanceBuddyException):
        fix_parser_dataframe.row_to_transaction(row)


def test_test_row_to_transaction_no_value(fix_parser_dataframe: DataframeParser):
    date = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    row: list[str] = [date.isoformat(), "EUR"]
    fix_parser_dataframe.config.settings.field_map = {
        ReportTransactionField.DATE: 0,
        ReportTransactionField.CURRENCY: 1,
    }
    transction = fix_parser_dataframe.row_to_transaction(row)
    assert transction.date == date
    assert transction.value == 0
    assert transction.display == "0.00"
    assert transction.currency == "EUR"



# package
from financebuddy.parserconfig.models import ParserConfig, ParserConfigSettings
from financebuddy.report.models import ReportTransactionField


class DataframeParserConfigSettings(ParserConfigSettings):
    field_map: dict[ReportTransactionField, int]
    date_format: str | None = None
    currency: str | None = None


class DataframeParserConfig(ParserConfig):
    settings: DataframeParserConfigSettings



# stdlib
import os

# package
from financebuddy import cli


def test_cli_missing_action(capfd):
    code = cli.run([])
    assert code == 1

    out, _ = capfd.readouterr()
    out = out.rstrip(os.linesep)

    assert out == "financebuddy-cli: error: missing or incomplete action, see -h/--help"



# package
from financebuddy.parserconfig.models import ParserConfig, ParserExtension
from financebuddy.report.models import Report


class Parser:
    extension: ParserExtension = ParserExtension.UNKNOWN

    def __init__(self, config: ParserConfig) -> None:
        self.config = config

    def generate_report(self, path: str) -> Report:
        raise NotImplementedError()



# third party
import pytest

# package
from financebuddy import currencyutils


def test_validate_currency():
    assert currencyutils.validate_currency("eur") == "EUR"


def test_validate_currency_raise():
    with pytest.raises(ValueError):
        currencyutils.validate_currency("???")


@pytest.mark.parametrize(
    ["currency", "expected"],
    [
        ["EUR", 2],
        ["KWD", 3],
        ["YER", 0],
    ],
)
def test_get_currency_precision(currency: str, expected: int):
    assert currencyutils.get_currency_precision(currency) == expected


def test_calculate_smallest_value():
    assert currencyutils.calculate_smallest_value(10, "EUR") == 1000


def test_calculate_currency_value():
    assert currencyutils.calculate_currency_value(1000, "EUR") == 10.00


def test_format_value():
    assert currencyutils.format_value(1000, "EUR") == "10.00"


def test_list_currencies():
    assert len(currencyutils.list_currencies()) > 1



# stdlib
import csv
import json
from typing import Any

# package
from financebuddy.export.integrations.base import Exporter
from financebuddy.export.models import ExporterExtension, ExporterFormat, ExportReport


class FinanceBuddyCSVExporter(Exporter):
    format: ExporterFormat = ExporterFormat.FINANCEBUDDY
    extension: ExporterExtension = ExporterExtension.CSV

    def dump_report(self, export_report: ExportReport) -> str:
        path = self.build_export_report_path()
        raw_report: dict = json.loads(export_report.model_dump_json())
        raw_transactions: list[dict[str, Any]] = raw_report.get("transactions", [dict()])
        headers = list(raw_transactions[0].keys())
        with open(path, "w") as fp:
            writer = csv.DictWriter(fp, headers)
            writer.writeheader()
            writer.writerows(raw_transactions)

        return path



# stdlib
import warnings

# third party
import pandas as pd

# package
from financebuddy.parser.integrations.dataframe import DataframeParser
from financebuddy.parserconfig.integrations.excel import ExcelParserConfig
from financebuddy.parserconfig.models import ParserExtension


class ExcelParser(DataframeParser):
    extension: ParserExtension = ParserExtension.XLSX

    def __init__(self, config: ExcelParserConfig) -> None:
        super().__init__(config)
        self.config = config

    def read_file(self, path: str) -> pd.DataFrame:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df = pd.read_excel(path, engine="openpyxl")
        return df



# stdlib
import os

# package
import financebuddy_parsers

# third party
import pytest


@pytest.fixture
def fix_dir_data() -> str:
    current_dir = os.path.dirname(__file__)
    data_dir = os.path.abspath(os.path.join(current_dir, "..", "data"))
    return data_dir


@pytest.fixture
def fix_financebuddy_parsers_dir_configs(fix_dir_data) -> str:
    dir_config = os.path.join(fix_dir_data, "configs")
    return dir_config


@pytest.fixture
def mock_financebuddy_parsers_dir_configs(fix_financebuddy_parsers_dir_configs, monkeypatch):
    monkeypatch.setattr(financebuddy_parsers, "DIR_CONFIGS", fix_financebuddy_parsers_dir_configs)



# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parser.integrations.base import Parser
from financebuddy.parser.integrations.csv import CSVParser
from financebuddy.parser.integrations.excel import ExcelParser
from financebuddy.parserconfig.models import ParserConfig, ParserExtension

MAP_PARSER_BY_EXT: dict[ParserExtension, type[Parser]] = {
    ParserExtension.CSV: CSVParser,
    ParserExtension.XLSX: ExcelParser,
}


def get_parser_type(extension: ParserExtension) -> type[Parser]:
    try:
        ParserType = MAP_PARSER_BY_EXT[extension]
    except KeyError:
        raise FinanceBuddyException(f"no parser found for extension: {extension}")
    return ParserType


def get_parser(config: ParserConfig) -> Parser:
    ParserType = get_parser_type(config.extension)
    parser = ParserType(config)
    return parser



# third party
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException


def test_init():
    with pytest.raises(FinanceBuddyException):
        raise FinanceBuddyException("hjelp")



# stdlib
from typing import Any

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parserconfig.integrations.csv import CSVParserConfig
from financebuddy.parserconfig.integrations.excel import ExcelParserConfig
from financebuddy.parserconfig.models import ParserConfig, ParserExtension

MAP_PARSER_CONFIGS_BY_EXT: dict[tuple[str | None, ParserExtension], type[ParserConfig]] = {
    (None, ParserExtension.CSV): CSVParserConfig,
    (None, ParserExtension.XLSX): ExcelParserConfig,
}


def get_parser_config_type(format: str | None, extension: ParserExtension) -> type[ParserConfig]:
    key = (format, extension)
    try:
        ParserConfigType = MAP_PARSER_CONFIGS_BY_EXT[key]
    except KeyError:
        msg = f"parser config not supported for format/extension: {format}/{extension}"
        raise FinanceBuddyException(msg)
    return ParserConfigType


def get_parser_config(
    path: str,
    format: str | None,
    extension: ParserExtension,
    raw_content: dict[str, Any],
) -> ParserConfig:
    ParserConfigType = get_parser_config_type(format, extension)
    parser_config = ParserConfigType(path=path, **raw_content)
    return parser_config



# third party
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.export import factory
from financebuddy.export.integrations.base import Exporter
from financebuddy.export.integrations.financebuddy_csv import FinanceBuddyCSVExporter
from financebuddy.export.integrations.financebuddy_json import FinanceBuddyJSONExporter
from financebuddy.export.models import ExporterExtension, ExporterFormat


@pytest.mark.parametrize(
    ["format", "extension", "ExporterType"],
    [
        (ExporterFormat.FINANCEBUDDY, ExporterExtension.JSON, FinanceBuddyJSONExporter),
        (ExporterFormat.FINANCEBUDDY, ExporterExtension.CSV, FinanceBuddyCSVExporter),
    ],
)
def test_get_exporter_type(
    format: ExporterFormat,
    extension: ExporterExtension,
    ExporterType: type[Exporter],
):
    assert factory.get_exporter_type(format, extension) == ExporterType


def test_get_exporter_type_raise_FinanceBuddyException():
    with pytest.raises(FinanceBuddyException):
        factory.get_exporter_type("?", "?")


def test_get_exporter():
    exporter = factory.get_exporter(ExporterFormat.FINANCEBUDDY, ExporterExtension.JSON)
    assert isinstance(exporter, FinanceBuddyJSONExporter)


def test_get_exporters():
    assert len(factory.get_exporters()) == 2



# stdlib
import datetime
import os
import tempfile

# package
from financebuddy.export.models import ExporterExtension, ExporterFormat, ExportReport, ExportTransaction
from financebuddy.report.models import Report


class Exporter:
    format: ExporterFormat = ExporterFormat.UNKNOWN
    extension: ExporterExtension = ExporterExtension.UNKNOWN

    @staticmethod
    def report_to_export_report(report: Report) -> ExportReport:
        export_transactions: list[ExportTransaction] = []
        for item in report.items:
            if transaction := item.transaction:
                export_transaction = ExportTransaction.model_validate(transaction.model_dump())
                export_transactions.append(export_transaction)
        export_report = ExportReport(transactions=export_transactions)
        return export_report

    def build_export_report_path(self) -> str:
        directory = tempfile.gettempdir()
        timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
        basename = f"financebuddy_export_{self.format}_{timestamp}.{self.extension}"
        path = os.path.join(directory, basename)
        return path

    def dump_report(self, export_report: ExportReport) -> str:
        raise NotImplementedError()

    def export_report(self, report: Report) -> str:
        export_report = self.report_to_export_report(report)
        path = self.dump_report(export_report)
        return path



# stdlib
import os

# third party
import pytest

# package
from financebuddy import cli
from financebuddy.exceptions import FinanceBuddyException


def test_cli_exception(capfd, monkeypatch):
    def mock_raise(*args, **kwargs):
        raise RuntimeError("crash")

    monkeypatch.setattr(cli, "list_parsers", mock_raise)

    with pytest.raises(RuntimeError):
        cli.run(["parsers", "list"])


def test_cli_exception_package(capfd, monkeypatch):
    def mock_raise(*args, **kwargs):
        raise FinanceBuddyException("oops!")

    monkeypatch.setattr(cli, "list_parsers", mock_raise)

    code = cli.run(["parsers", "list"])
    assert code == 1

    out, _ = capfd.readouterr()
    out = out.rstrip(os.linesep)

    assert out == "financebuddy-cli: error: oops!"



# stdlib
import json
import os
from typing import Any

# third party
import financebuddy_parsers

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parserconfig import factory
from financebuddy.parserconfig.models import ParserConfig, ParserExtension


def load_parser_config(path: str) -> ParserConfig:
    if not os.path.exists(path):
        raise FinanceBuddyException(f"parser config file does not exist: {path}")

    try:
        with open(path, "r") as fp:
            content: dict[str, Any] = json.load(fp)
    except Exception:
        raise FinanceBuddyException(f"parser config file does not contain a valid JSON object: {path}")

    raw_extension = content.get("extension")
    if not raw_extension:
        raise FinanceBuddyException(f"parser config file does not contain key 'extension': {path}")

    extension = ParserExtension(raw_extension)
    if extension == ParserExtension.UNKNOWN:
        raise FinanceBuddyException(f"file contains an unknown extension '{raw_extension}': {path}")

    format: str | None = None
    if extension not in [ParserExtension.CSV, ParserExtension.XLSX]:
        format = content.get("format")

    configuration = factory.get_parser_config(path, format, extension, content)
    return configuration


def get_parser_configs() -> list[ParserConfig]:
    configs = list(map(load_parser_config, financebuddy_parsers.PARSER_CONFIGS))
    return configs


def build_path(format: str, extension: str) -> str:
    basename = f"parser_{format}_{extension}_config.json"
    path = os.path.join(financebuddy_parsers.DIR_CONFIGS, basename)
    return path


def find_parser_config(format: str, extension: str) -> ParserConfig:
    path = build_path(format, extension)
    try:
        config = load_parser_config(path)
        return config
    except FinanceBuddyException:
        raise FinanceBuddyException(f"no parser found for format/extension: {format}/{extension}")



# stdlib
from argparse import ArgumentParser

# third party
from tabulate import tabulate

# package
from financebuddy import __version__
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.export import api as exportapi
from financebuddy.parser import api as parserapi
from financebuddy.parserconfig import api as configapi
from financebuddy.report import api as reportapi

OUT_ERROR_PREFIX = "financebuddy-cli: error:"


# ==============================================================================
# actions
# ==============================================================================
def list_parsers() -> None:
    configs = configapi.get_parser_configs()
    table_data = [[config.format, config.extension] for config in configs]
    print(tabulate(table_data, headers=["format", "extension"], tablefmt="rounded_grid"))


def parse_file(format: str, extension: str, input: str) -> None:
    config = configapi.find_parser_config(format, extension)
    report = parserapi.generate_report(input, config)
    report_path = reportapi.dump_report(report)
    print(report_path)


def list_exporters() -> None:
    exporter_types = exportapi.get_exporters()
    table_data = [[Exporter.format, Exporter.extension] for Exporter in exporter_types]
    print(tabulate(table_data, headers=["format", "extension"], tablefmt="rounded_grid"))


def export_report(format: str, extension: str, input: str) -> None:
    report = reportapi.load_report(input)
    export_path = exportapi.export_report(report, format, extension)
    print(export_path)


# ==============================================================================
# parser
# ==============================================================================
def build_parser() -> ArgumentParser:
    description = "FinanceBuddy is a tool that centralizes and parses data from diffent banks into a unified format, streamlining the process for analytics and reporting."
    parser = ArgumentParser(description=description)
    subparsers = parser.add_subparsers()

    # root arguments
    parser.add_argument("-v", "--version", action="store_true", help="print the release version")

    # parsers
    p_parsers = subparsers.add_parser("parsers")
    p_parsers_sub = p_parsers.add_subparsers()

    # parsers - list
    p_parse_list = p_parsers_sub.add_parser("list", help="list the available parsers")
    p_parse_list.set_defaults(func=list_parsers)

    # parsers - parse
    p_parse = p_parsers_sub.add_parser("parse", help="parses a bank file to a report file")
    p_parse.add_argument("-f", "--format", required=True, help="format of the bank file")
    p_parse.add_argument("-e", "--extension", required=True, help="extension of the bank file")
    p_parse.add_argument("-i", "--input", required=True, help="path of the bank file")
    p_parse.set_defaults(func=parse_file)

    # exporters
    p_exporters = subparsers.add_parser("exporters")
    p_exporters_sub = p_exporters.add_subparsers()

    # export - list
    p_export_list = p_exporters_sub.add_parser("list", help="list the available exporters")
    p_export_list.set_defaults(func=list_exporters)

    # exporters - export
    p_export = p_exporters_sub.add_parser("export", help="exports a report file")
    p_export.add_argument("-f", "--format", required=True, help="format of the export file")
    p_export.add_argument("-e", "--extension", required=True, help="extension of the export file")
    p_export.add_argument("-i", "--input", required=True, help="path of the report file")
    p_export.set_defaults(func=export_report)

    return parser


# ==============================================================================
# main
# ==============================================================================
def run(args: list[str] | None = None) -> int:
    parser = build_parser()
    namespace = parser.parse_args(args)
    kwargs = vars(namespace)

    version = kwargs.pop("version")
    if version:
        print(__version__)
        return 0

    try:
        func = kwargs.pop("func")
    except KeyError:
        print(f"{OUT_ERROR_PREFIX} missing or incomplete action, see -h/--help")
        return 1

    try:
        func(**kwargs)
    except FinanceBuddyException as e:
        print(f"{OUT_ERROR_PREFIX} {e}")
        return 1

    return 0


if __name__ == "__main__":
    code = run()
    exit(code)



# third party
import babel.numbers
import pycountry


def validate_currency(currency: str) -> str:
    code = pycountry.currencies.get(alpha_3=currency)
    if code is None:
        raise ValueError(f"unknown currency: {currency}")
    return code.alpha_3


def get_currency_precision(currency: str) -> int:
    format_str = babel.numbers.format_currency(1, currency, locale="en_US", format="#,##0.00")

    parts = format_str.split(".")
    try:
        _, digits = parts
    except ValueError:
        digits = ""

    precision = len(digits)
    return precision


def calculate_smallest_value(value: int | float, currency: str) -> int:
    precision = get_currency_precision(currency)
    smallest_value = int(value * (10**precision))
    return smallest_value


def calculate_currency_value(value: int, currency: str) -> float:
    precision = get_currency_precision(currency)
    currency_value = float(value / (10**precision))
    return currency_value


def format_value(value: int, currency: str) -> str:
    precision = get_currency_precision(currency)
    display_value = value / (10**precision)
    formatted = babel.numbers.format_currency(display_value, currency, locale="en_US", format="#,##0.00")
    return formatted


def list_currencies() -> list[str]:
    currencies = list(babel.numbers.list_currencies())
    currencies.sort()
    return currencies



# third party
import pandas as pd

# package
from financebuddy.parser.integrations.dataframe import DataframeParser
from financebuddy.parserconfig.integrations.csv import CSVParserConfig
from financebuddy.parserconfig.models import ParserExtension


class CSVParser(DataframeParser):
    extension: ParserExtension = ParserExtension.CSV

    def __init__(self, config: CSVParserConfig) -> None:
        super().__init__(config)
        self.config = config

    def read_file(self, path: str) -> pd.DataFrame:
        df = pd.read_csv(path, sep=self.config.settings.delimiter)
        return df



# stdlib
import os

# third party
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parserconfig import api


@pytest.mark.parametrize(
    ["basename", "format", "extension"],
    [
        ("parser_foo_csv_config.json", "foo", "csv"),
    ],
)
def test_load_parser_config(
    basename: str,
    format: str,
    extension: str,
    fix_financebuddy_parsers_dir_configs: str,
):
    path = os.path.join(fix_financebuddy_parsers_dir_configs, basename)
    config = api.load_parser_config(path)
    assert config.format == format
    assert config.extension == extension


def test_load_parser_config_fail_path_does_not_exist():
    with pytest.raises(FinanceBuddyException):
        api.load_parser_config("")


@pytest.mark.parametrize(
    "basename",
    [
        "not_a_json.txt",
        "missing_extension.json",
        "invalid_extension.json",
    ],
)
def test_load_parser_config_fail(
    basename: str,
    fix_financebuddy_parsers_dir_configs: str,
    mock_financebuddy_parsers_dir_configs: None,
):
    path = os.path.join(fix_financebuddy_parsers_dir_configs, basename)
    with pytest.raises(FinanceBuddyException):
        api.load_parser_config(path)


def test_build_path(
    fix_financebuddy_parsers_dir_configs: str,
    mock_financebuddy_parsers_dir_configs: None,
):
    path = api.build_path("unknown", "unknown")
    dirname, basename = os.path.split(path)
    assert dirname == fix_financebuddy_parsers_dir_configs
    assert basename == "parser_unknown_unknown_config.json"


def test_find_parser_config(mock_financebuddy_parsers_dir_configs: None):
    config = api.find_parser_config("foo", "csv")
    assert config.format == "foo"
    assert config.extension == "csv"


def test_find_parser_config_fail():
    with pytest.raises(FinanceBuddyException):
        api.find_parser_config("?", "?")



# stdlib
import hashlib
import re

REGEX_WHITESPACE = re.compile(r"\s+")


def remove_whitespace(s: str) -> str:
    result = re.sub(REGEX_WHITESPACE, "", s)
    return result


def remove_plus(s: str) -> str:
    result = s.replace("+", "")
    return result


def sanitize_value_string(s: str) -> str:
    result = remove_whitespace(s)
    result = remove_plus(result)

    dot_pos = result.rfind(".")
    comma_pos = result.rfind(",")
    if comma_pos > dot_pos:
        result = result.replace(".", "")
        result = result.replace(",", ".")
    else:
        result = result.replace(",", "")
    return result


def hash_string(s: str) -> str:
    h = hashlib.md5(s.encode("utf-8")).hexdigest()
    return h



# stdlib
import datetime
import enum

# third party
from pydantic import BaseModel


class ReportTransactionField(enum.StrEnum):
    DATE = "date"
    VALUE = "value"
    CURRENCY = "currency"
    DISPLAY = "display"
    SRC_BANK_NAME = "src_bank_name"
    SRC_ACCOUNT_HOLDER = "src_account_holder"
    SRC_ACCOUNT_NUMBER = "src_account_number"
    DST_BANK_NAME = "dst_bank_name"
    DST_ACCOUNT_HOLDER = "dst_account_holder"
    DST_ACCOUNT_NUMBER = "dst_account_number"
    DESCRIPTION = "description"


class ReportTransaction(BaseModel):
    src_bank_name: str
    src_account_holder: str | None = None
    src_account_number: str | None = None
    dst_bank_name: str | None = None
    dst_account_holder: str | None = None
    dst_account_number: str | None = None
    date: datetime.datetime
    value: int
    display: str
    currency: str
    description: str | None = None
    raw: str
    hash: str


class ReportError(BaseModel):
    raw: str
    message: str
    exception: str


class ReportItem(BaseModel):
    transaction: ReportTransaction | None = None
    error: ReportError | None = None


class ReportSummary(BaseModel):
    total: int
    parsed: int
    failed: int


class Report(BaseModel):
    items: list[ReportItem]
    summary: ReportSummary



# stdlib
import datetime

# third party
import pytest

# package
from financebuddy.report.models import Report, ReportError, ReportItem, ReportSummary, ReportTransaction


@pytest.fixture
def fix_report_transaction() -> ReportTransaction:
    transaction = ReportTransaction(
        src_bank_name="bank1",
        src_account_holder="bank1_holder1",
        src_account_number="bank1_holder1_account_1",
        dst_bank_name="bank2",
        dst_account_holder="bank2_holder1",
        dst_account_number="bank2_holder1_account_1",
        date=datetime.datetime.now(tz=datetime.timezone.utc),
        value=100,
        currency="EUR",
        display="1.00",
        description="foo",
        raw="[]",
        hash="foo",
    )
    return transaction


@pytest.fixture
def fix_report_error() -> ReportError:
    error = ReportError(raw="[]", message="", exception="")
    return error


@pytest.fixture
def fix_report(fix_report_transaction, fix_report_error) -> Report:
    report = Report(
        items=[
            ReportItem(transaction=fix_report_transaction),
            ReportItem(error=fix_report_error),
        ],
        summary=ReportSummary(
            total=2,
            parsed=1,
            failed=1,
        ),
    )
    return report



# third party
import pytest

# package
from financebuddy.parser.integrations.base import Parser
from financebuddy.parserconfig.models import ParserConfig


def test_init(fix_parser_config: ParserConfig):
    assert Parser.extension == "unknown"
    parser = Parser(fix_parser_config)
    assert parser.config.format == "foo"
    assert parser.config.extension == "unknown"


def test_generate_report(fix_parser: Parser):
    with pytest.raises(NotImplementedError):
        fix_parser.generate_report("/tmp/foo.csv")



# stdlib
import datetime
import os

# third party
import pytest
from financebuddy_parsers import MAP_SAMPLE_FILES

# package
from financebuddy import cli
from financebuddy.report import api as reportapi

values = []
for k, v in MAP_SAMPLE_FILES.items():
    format, extension = k
    for path in v:
        values.append((format, extension, path))


@pytest.mark.parametrize(["format", "extension", "path"], values)
def test_parse_file(format: str, extension: str, path: str, capfd):
    code = cli.run(["parsers", "parse", "-f", format, "-e", extension, "-i", path])
    assert code == 0

    out, _ = capfd.readouterr()
    out = out.rstrip(os.linesep)

    report = reportapi.load_report(out)
    assert report.summary.parsed == 3
    assert report.summary.failed == 0

    i0 = report.items[0]
    assert i0.error is None
    assert i0.transaction is not None
    t0 = i0.transaction
    assert t0.src_bank_name == format
    assert t0.date == datetime.datetime(1970, 1, 1)
    assert t0.currency == "EUR"
    assert t0.value == -100000

    i1 = report.items[1]
    assert i1.error is None
    assert i1.transaction is not None
    t1 = i1.transaction
    assert t1.src_bank_name == format
    assert t1.date == datetime.datetime(1970, 1, 1)
    assert t1.currency == "EUR"
    assert t1.value == 0

    i2 = report.items[2]
    assert i2.error is None
    assert i2.transaction is not None
    t2 = i2.transaction
    assert t2.src_bank_name == format
    assert t2.date == datetime.datetime(1970, 1, 1)
    assert t2.currency == "EUR"
    assert t2.value == 100000



# stdlib
import os

# package
from financebuddy import __version__, cli


def test_cli_version(capfd):
    code = cli.run(["--version"])
    assert code == 0

    out, _ = capfd.readouterr()
    out = out.rstrip(os.linesep)

    assert out == __version__



# package
from financebuddy.parserconfig.integrations.dataframe import DataframeParserConfig, DataframeParserConfigSettings
from financebuddy.parserconfig.models import ParserExtension


class ExcelParserConfigSettings(DataframeParserConfigSettings):
    pass


class ExcelParserConfig(DataframeParserConfig):
    type: ParserExtension = ParserExtension.XLSX
    settings: ExcelParserConfigSettings



# stdlib
import datetime
import decimal
import json

# third party
import dateutil.parser
import pandas as pd

# package
from financebuddy import currencyutils
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parser import helpers
from financebuddy.parser.integrations.base import Parser
from financebuddy.parserconfig.integrations.dataframe import DataframeParserConfig
from financebuddy.report.models import (
    Report,
    ReportError,
    ReportItem,
    ReportSummary,
    ReportTransaction,
    ReportTransactionField,
)


def pandas_serialize_json(v):
    if isinstance(v, pd.Timestamp):
        v = v.to_pydatetime().isoformat()
    elif isinstance(v, (int, float)):
        v = str(v)
    return v


class DataframeParser(Parser):
    def __init__(self, config: DataframeParserConfig) -> None:
        super().__init__(config)
        self.config = config

    def read_file(self, path: str) -> pd.DataFrame:
        raise NotImplementedError()

    def row_to_transaction(self, row: list[str]) -> ReportTransaction:
        raw_transaction: dict = {"src_bank_name": self.config.format}

        # raw
        raw_str = json.dumps(row)
        raw_transaction["raw"] = raw_str

        # hash
        raw_hash = helpers.hash_string(raw_str)
        raw_transaction["hash"] = raw_hash

        # field map parsing
        for field, column in self.config.settings.field_map.items():
            v = row[column]

            if field in [
                ReportTransactionField.SRC_ACCOUNT_NUMBER,
                ReportTransactionField.DST_ACCOUNT_NUMBER,
            ]:
                v = helpers.remove_whitespace(v)
            elif field == ReportTransactionField.DATE:
                if date_format := self.config.settings.date_format:
                    dt = datetime.datetime.strptime(v, date_format)
                else:
                    dt = dateutil.parser.parse(v)
                v = dt
            elif field == ReportTransactionField.VALUE:
                v = helpers.sanitize_value_string(v)
                v = decimal.Decimal(v)
            elif field == ReportTransactionField.CURRENCY:
                code = currencyutils.validate_currency(v)
                v = code

            raw_transaction[field] = v

        # currency
        currency = raw_transaction.get(ReportTransactionField.CURRENCY, self.config.settings.currency)
        if currency is None:
            raise FinanceBuddyException("no currency could be determined")

        # value
        value = raw_transaction.get(ReportTransactionField.VALUE, decimal.Decimal(0))
        value = currencyutils.calculate_smallest_value(value, currency)
        raw_transaction[ReportTransactionField.VALUE] = value

        # display
        display = currencyutils.format_value(value, currency)
        raw_transaction[ReportTransactionField.DISPLAY] = display

        transcation = ReportTransaction(**raw_transaction)
        return transcation

    def dataframe_to_items(self, df: pd.DataFrame) -> list[ReportItem]:
        df.fillna("", inplace=True)
        items: list[ReportItem] = []

        for t in df.itertuples(index=False):
            row: list[str] = list(map(pandas_serialize_json, t))

            transaction: ReportTransaction | None = None
            error: ReportError | None = None

            try:
                transaction = self.row_to_transaction(row)
            except Exception as e:
                error = ReportError(
                    raw=json.dumps(row),
                    message="",
                    exception=str(e),
                )

            item = ReportItem(transaction=transaction, error=error)
            items.append(item)

        return items

    def generate_report(self, path: str) -> Report:
        df = self.read_file(path)
        items = self.dataframe_to_items(df)

        nr_of_transactions = 0
        nr_of_errors = 0
        for item in items:
            if item.error:
                nr_of_errors += 1
            elif item.transaction:
                nr_of_transactions += 1

        report = Report(
            items=items,
            summary=ReportSummary(
                total=df.shape[0],
                parsed=nr_of_transactions,
                failed=nr_of_errors,
            ),
        )
        return report



# third party
import pytest

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.parserconfig import factory
from financebuddy.parserconfig.integrations.csv import CSVParserConfig
from financebuddy.parserconfig.models import ParserConfig, ParserExtension


@pytest.mark.parametrize(
    ["format", "extension", "ParserConfigType"],
    [
        (None, ParserExtension.CSV, CSVParserConfig),
    ],
)
def test_get_parser_config_type(
    format: str | None,
    extension: ParserExtension,
    ParserConfigType: type[ParserConfig],
):
    assert factory.get_parser_config_type(format, extension) == ParserConfigType


def test_get_parser_config_type_raise_FinanceBuddyException():
    with pytest.raises(FinanceBuddyException):
        factory.get_parser_config_type("?", "?")


def test_get_parser_config():
    config = factory.get_parser_config(
        "/tmp/foo.json",
        None,
        ParserExtension.CSV,
        raw_content={
            "format": "foo",
            "extension": "csv",
            "settings": {"field_map": {}},
        },
    )
    assert isinstance(config, CSVParserConfig)



# package
from financebuddy.export import factory
from financebuddy.export.integrations.base import Exporter
from financebuddy.export.models import ExporterExtension, ExporterFormat
from financebuddy.report.models import Report


def get_exporters() -> list[type[Exporter]]:
    return factory.get_exporters()


def export_report(report: Report, format: ExporterFormat, extension: ExporterExtension) -> str:
    exporter = factory.get_exporter(format, extension)
    path = exporter.export_report(report)
    return path



# package
from financebuddy.parser import factory
from financebuddy.parserconfig.models import ParserConfig
from financebuddy.report.models import Report


def generate_report(path: str, config: ParserConfig) -> Report:
    parser = factory.get_parser(config)
    report = parser.generate_report(path)
    return report



# third party
import pytest

# package
from financebuddy.parser.integrations.base import Parser
from financebuddy.parser.integrations.dataframe import DataframeParser
from financebuddy.parserconfig.integrations.dataframe import DataframeParserConfig, DataframeParserConfigSettings
from financebuddy.parserconfig.models import ParserConfig, ParserExtension


@pytest.fixture
def fix_parser_config() -> ParserConfig:
    config = ParserConfig(
        path="/tmp/foo.json",
        format="foo",
        extension=ParserExtension.UNKNOWN,
    )
    return config


@pytest.fixture
def fix_parser(fix_parser_config: ParserConfig) -> Parser:
    parser = Parser(fix_parser_config)
    return parser


@pytest.fixture
def fix_parser_config_dataframe() -> DataframeParserConfig:
    config = DataframeParserConfig(
        path="/tmp/foo.json",
        format="foo",
        extension=ParserExtension.UNKNOWN,
        settings=DataframeParserConfigSettings(field_map={}),
    )
    return config


@pytest.fixture
def fix_parser_dataframe(fix_parser_config_dataframe: DataframeParserConfig) -> DataframeParser:
    parser = DataframeParser(fix_parser_config_dataframe)
    return parser



# stdlib
from copy import deepcopy

# package
from financebuddy.exceptions import FinanceBuddyException
from financebuddy.export.integrations.base import Exporter
from financebuddy.export.integrations.financebuddy_csv import FinanceBuddyCSVExporter
from financebuddy.export.integrations.financebuddy_json import FinanceBuddyJSONExporter
from financebuddy.export.models import ExporterExtension, ExporterFormat

EXPORTERS = {
    (ExporterFormat.FINANCEBUDDY, ExporterExtension.CSV): FinanceBuddyCSVExporter,
    (ExporterFormat.FINANCEBUDDY, ExporterExtension.JSON): FinanceBuddyJSONExporter,
}


def get_exporter_type(format: ExporterFormat, extension: ExporterExtension) -> type[Exporter]:
    key = (format, extension)
    try:
        ExporterType = EXPORTERS[key]
    except KeyError:
        msg = f"no exporter found for format/extension: {format}/{extension}"
        raise FinanceBuddyException(msg)
    return ExporterType


def get_exporter(format: ExporterFormat, extension: ExporterExtension) -> Exporter:
    ExporterType = get_exporter_type(format, extension)
    exporter = ExporterType()
    return exporter


def get_exporters() -> list[type[Exporter]]:
    exporter_types = list(EXPORTERS.values())
    return deepcopy(exporter_types)



# stdlib
from enum import StrEnum

# third party
from pydantic import BaseModel


class ParserExtension(StrEnum):
    UNKNOWN = "unknown"
    CSV = "csv"
    XLSX = "xlsx"

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


class ParserConfigSettings(BaseModel):
    pass


class ParserConfig(BaseModel):
    path: str
    format: str
    extension: ParserExtension
    settings: ParserConfigSettings | None = None



# third party
import pytest

# package
from financebuddy.parser import helpers


@pytest.mark.parametrize(
    ["s", "expected"],
    [
        ("", ""),
        (" ", ""),
        ("foo", "foo"),
        ("foo ", "foo"),
        (" foo", "foo"),
        (" foo ", "foo"),
    ],
)
def test_remove_whitespace(s: str, expected: str):
    assert helpers.remove_whitespace(s) == expected


@pytest.mark.parametrize(
    ["s", "expected"],
    [
        ("", ""),
        ("+", ""),
        ("-", "-"),
    ],
)
def test_remove_plus(s: str, expected: str):
    assert helpers.remove_plus(s) == expected


@pytest.mark.parametrize(
    ["inp", "out"],
    [
        # single digit
        ("0", "0"),
        ("1", "1"),
        # single decimal
        ("1,0", "1.0"),
        # multiple decimal
        ("1,00", "1.00"),
        # financial notation
        ("1,000.00", "1000.00"),
        ("1,000,000.00", "1000000.00"),
        # signed value
        ("+50.00", "50.00"),
        ("-50.00", "-50.00"),
    ],
)
def test_sanitize_value_string(inp: str, out: str):
    assert helpers.sanitize_value_string(inp) == out


def test_hash_string():
    assert helpers.hash_string("0") == "cfcd208495d565ef66e7dff9f98764da"
